#!/bin/sh

JAR=~anders/anna-20120702.jar
TRAIN=all_train_v4_auto_conll.c09
DEV=all_dev_v4_auto_conll.c09

COUNT="-count 2000"
#LEMMA_UC=""
LEMMA_UC="-uc"


for i in `seq 0 9`; do
    java -cp $JAR is2.lemmatizer.Lemmatizer $LEMMA_UC -train split/p-${i} -test split/r-${i} -model split/l-${i}.mdl -out split/l-${i} -eval split/r-${i} $COUNT &> split/l-${i}.log &
done

java -cp $JAR is2.lemmatizer.Lemmatizer $LEMMA_UC -train $TRAIN -test $DEV -model all_lemma.mdl -out ${DEV}.l -eval $DEV $COUNT &> all_lemma.log &

echo "Waiting for lemmatizer jobs to finish..."
wait
echo "Done waiting!"
sync

exit 0;

echo "Preparing tagger training data"

echo -n "" > ${TRAIN}.l
for i in `seq 0 9`; do
    echo -n "" > split/tagtrain-${i}
    cat split/l-${i} >> ${TRAIN}.l
    for j in `seq 0 9`; do
	if [ $i -ne $j ]; then
#	    echo "$j -> $i"
	    cat split/l-${j} >> split/tagtrain-${i}
	fi
    done
done

sync

echo "Training tagger"

for i in `seq 0 9`; do
    java -cp $JAR is2.tag.Tagger -train split/tagtrain-${i} -test split/l-${i} -model split/t-${i}.mdl -out split/t-${i} -eval split/r-${i} $COUNT &> split/t-${i}.log &
done

java -cp $JAR is2.tag.Tagger -train ${TRAIN}.l -test ${DEV}.l -model all_tag.mdl -out ${DEV}.lt -eval ${DEV} $COUNT &> all_tag.log &

echo "Waiting for jobs to finish..."
wait
echo "Done."

sync

echo "Preparing parser training data"

echo -n "" > ${TRAIN}.lt
for i in `seq 0 9`; do
    echo -n "" > split/parsetrain-${i}
    cat split/t-${i} >> ${TRAIN}.lt
    for j in `seq 0 9`; do
	if [ $i -ne $j ]; then
#	    echo "$j -> $i"
	    cat split/t-${j} >> split/parsetrain-${i}
	fi
    done
done

sync

for i in `seq 0 5`; do
    java -cp $JAR is2.parser.Parser -train split/parsetrain-${i} -test split/t-${i} -model split/prs-${i}.mdl -out split/prs-${i} -eval split/t-${i} -cores 3 $COUNT &> split/prs-${i}.log &
done

echo "Training first batch of parsers, waiting..."
wait
echo "Done"


for i in `seq 6 9`; do
    java -cp $JAR is2.parser.Parser -train split/parsetrain-${i} -test split/t-${i} -model split/prs-${i}.mdl -out split/prs-${i} -eval split/t-${i} -cores 3 $COUNT &> split/prs-${i}.log &
done

java -cp $JAR is2.parser.Parser -train ${TRAIN}.lt -test ${DEV}.lt -model all_parse.mdl -out ${DEV}.ltp -eval ${DEV} -cores 4 $COUNT &> all_parse.log &

echo "Training second batch of parsers, waiting..."
wait
echo "Done"

echo "Concatenating train data"
echo -n "" > ${TRAIN}.ltp
for i in `seq 0 9`; do
    cat split/prs-${i} >> ${TRAIN}.ltp
done


