#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)';
binmode STDIN,':encoding(utf8)';
binmode STDOUT,':encoding(utf8)';


my $FIND_CONLL_DOCS_SCRIPT="./extract_doc_names_from_conll.sh";

die("Not enough arguments. Check the code for help") if(scalar(@ARGV)<3);

my $conllSetRoot=shift; #This should be something like ..../conll-2011/v2/data/train
my $convRoot=shift; #This is the root of the conversion, something like root/train
my $suffix=shift; #something like v4_auto_conll

my @p=`$FIND_CONLL_DOCS_SCRIPT $conllSetRoot $suffix`;

my %ALLOWED_FORMS_DEVIATIONS= ( "-LRB-" => "(", "-RRB-" => ")", "-LCB-" => "{", "-RCB-" => "}", "-LSB-" => "[", "-RSB-" => "]" );

my $fileCount=0;
foreach my $prefix (@p){
    chomp $prefix;
    print "$prefix\n";
    $fileCount++;
    my $conllFile=$conllSetRoot . $prefix . "." . $suffix;
    my $shortPrefix;
    if($prefix=~/.*(\/(?:bc|bn|mz|nw|pt|tc|wb)\/.*)/){
	$shortPrefix=$1;
    } else {
	die ("Failed to parse short prefix from $prefix");
    }
    my $depsFile=$convRoot . $shortPrefix  . "." . $suffix . ".c09";
    die ("CoNLL file $conllFile does not exist!") if(! -e $conllFile);
    die ("Deps file $depsFile does not exist!") if (! -e $depsFile);
    my $outFile=$convRoot . $shortPrefix . "." . $suffix . ".coref";
    sanityCheckAndOutput($conllFile,$depsFile,$outFile);
#    last if($fileCount==1);
}

sub sanityCheckAndOutput {
    my $conll=shift;
    my $deps=shift;
    my $out=shift;
    open my $C,$conll or die("Failed to open $conll");
    open my $D,$deps or die("Failed to open $deps");
    open OF,">$out" or die("Failed to open $out");
    my @co;
#    my @co=readSentence($C);
    while((@co=readSentence($C))){
	if($co[0]->[0]=~/^\#(begin|end) document/){
	    print OF $co[0]->[0];
#	    if($1 eq 'begin')
	    next;
	}
	my @de=readSentence($D);
	my @conllForms=map { $_->[3] } @co;
#    print join(",",@conllForms)," X \n";
	my @depsForms=map {$_->[1] } @de;
#    print join(",",@depsForms)," Y \n";

	sameForms(\@conllForms,\@depsForms) || die("not same forms");

	my $cfgCol=5;
	my $srlSenseCol=7;
	my $speakerCol=9;
	my $nerCol=10;
	my $srlArgColBegin=11;
	my $srlArgColCount=scalar(@{$co[0]})-12;
	my $corefCol=scalar(@{$co[0]})-1;
	for(my $i=0;$i<scalar(@co);++$i){
	    #remove the two srl cols from the deps
	    pop @{$de[$i]};
	    pop @{$de[$i]};
	    #Push them back on (from the conll12 file)
	    my $isPred="_";
	    my $sense="_";
	    if($co[$i]->[$srlSenseCol]!~/^(-|_)$/){
		$isPred="Y";
		$sense=$co[$i]->[$srlSenseCol];
	    }
	    push @{$de[$i]},$isPred;
	    push @{$de[$i]},$sense;
	    push @{$de[$i]},$co[$i]->[$cfgCol];
	    push @{$de[$i]},$co[$i]->[$nerCol];
	    push @{$de[$i]},$co[$i]->[$speakerCol];
	    for(my $j=0;$j<$srlArgColCount;++$j){
		push @{$de[$i]},$co[$i]->[$srlArgColBegin+$j];
	    }
	    push @{$de[$i]},$co[$i]->[$corefCol];
	}
	foreach(@de){
	    print OF join("\t",@{$_}),"\n";
	}
	print OF "\n";
    }

    ## Make a final sanity check and make sure the output and the
    ## input conll have the same number of lines.
    my ($x)=split(/\s+/,`wc -l $conll`);
    my ($y)=split(/\s+/,`wc -l $out`);
    if($x!=$y){
	print $x,"\n";
	print $y,"\n";
	die("Not same number of lines in CoNLL and output!");
    }
    close $C;
    close $D;
    close OF;
}

sub readSentence {
    my $fh=shift;
    my @lines;
    while(1){
        defined(my $line=readline($fh)) or return;
	if($line=~/^\#/){
	    push @lines,[$line];
	    return @lines;
	}
        chomp $line;
        last if($line=~/^\s*$/);
        my @cols=split(/\s+/,$line);
        push @lines,\@cols;
    }
    return @lines;
}

sub sameForms {
    my @a=@{(shift)};
    my @b=@{(shift)};
    return 0 if(scalar(@a)!=scalar(@b));
    for(my $i=0;$i<scalar(@a);++$i){
	if($a[$i] ne $b[$i]){
	    my $q=$a[$i];
	    $q=~s/-LRB-/\(/g;
	    $q=~s/-RRB-/\)/g;
	    if($q ne $b[$i] && $ALLOWED_FORMS_DEVIATIONS{$a[$i]} ne $b[$i]){
		print "'$a[$i]' -- $b[$i]\n";
		return 0;
	    }
	}
    }
    return 1;
}
