#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)'; #Equivalent to perl -CD,
binmode STDERR,":encoding(utf8)";
binmode STDOUT,":encoding(utf8)"; #Equivalent to -CS
binmode STDIN,":encoding(utf8)";

my $conllIN=shift;
my $augIN=shift;

open CONLL,$conllIN or die("Failed ot open $conllIN");
open AUG,$augIN or die("Failed to open $augIN");

my $lineCount=0;
my $senCount=0;
#my $conll;
while(my $conll=<CONLL>){
    $lineCount++;
    chomp $conll;
    my $aug=<AUG>;
    chomp $aug;
    if($conll=~/^\s*$/ && $aug=~/^\s*$/){
	$senCount++;
	next;
    }
    if($conll=~/^#/){
	die("Not same doc boundary at line $lineCount") if($conll ne $aug);
	next;
    }
    my $err;
    if($err=notSame($conll,$aug)){
	die($err . " at line " . $lineCount);
    }
}

print "Lines: $lineCount\n";
print "Sents: $senCount\n";

close CONLL;
close AUG;

sub notSame{
    my $conll=shift;
    my $aug=shift;
    my @c=split(/\s+/,$conll);
    my @a=split(/\s+/,$aug);
    return "Not correct offset in column count" if((scalar(@c)+2)!=scalar(@a));
    return "Not same DocumentID col [col 1]" if($c[0] ne $a[0]);
    return "Not same part no [col 2]" if($c[1] ne $a[1]);
    return "Not same word no [col 3]" if($c[2] ne $a[2]);
    return "Not same form [col 4]" if($c[3] ne $a[3]);
    return "Not same pos [col 5]" if($c[4] ne $a[4]);
    return "Not same parse bit [col 6]" if($c[5] ne $a[5]);
    return "Not same lemma [col 7]" if($c[6] ne $a[6]);
    return "Not same frameset id [col 8]" if($c[7] ne $a[7]);
    return "Not same word sense [col 9]" if($c[8] ne $a[8]);
    return "Not same speaker [col 10]" if($c[9] ne $a[9]);
    return "Not same NE [col 11]" if($c[10] ne $a[10]);
    #Check the SRL cols
    my $cBeg=11;
    my $aBeg=13; #aug has two columns of head/deprel
    my $srlCols=scalar(@c)-1-$cBeg;
    for(my $i=0;$i<$srlCols;++$i){
	return "Not same SRL arg in cols " . ($cBeg+$i) . " (CoNLL), " . ($aBeg+$i) . " (AUG)" if($c[$cBeg+$i] ne $a[$aBeg+$i]);
    }
    return "Not same ... [last col]" if($c[-1] ne $a[-1]);
    return undef;
}
