#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)'; #Equivalent to perl -CD,
binmode STDERR,":encoding(utf8)";
binmode STDOUT,":encoding(utf8)"; #Equivalent to -CS
binmode STDIN,":encoding(utf8)";

use Coref;

my $goldpronouns=shift;
my $sys=shift;

open GOLD,$goldpronouns or die("Failed to open gold pronouns");
open SYS,$sys or die("Failed to open sys");


my $TOTAL=0;
my $CORRECT=0;

my %DOC_TOTAL;
my %DOC_CORRECT;

my $GOLD_LINE;
my ($sysdoc,$hashref,$inv)=getNextSys();
while($GOLD_LINE=<GOLD>){
    chomp $GOLD_LINE;
    next if($GOLD_LINE=~/Y$/); #Ignore cataphors in the gold.
    if(!defined($sysdoc)){
	print "reached eof for sys\n";
	die("!");
    }
    my @Q=split(/\t/,$GOLD_LINE);
    my ($doc,$line,$form,$sent,$hasAnt)=splice(@Q,0,5);
    $TOTAL++;
    $DOC_TOTAL{$doc}++;
    if($doc ne $sysdoc){
	while(1){
#	    print "$sysdoc\n";
	    ($sysdoc,$hashref,$inv)=getNextSys();
	    die("Reached eof for sysdoc before finished") if(!defined($sysdoc));
	    last if($sysdoc eq $doc);
	}
    }
    my @left=getClosestToTheLeft($line,$inv,$hashref);
#    checkCorrect($line,\@Q,$hashref,$inv);
    if(isCorrect(\@left,\@Q)){
	$CORRECT++;
	$DOC_CORRECT{$doc}++;
    }
}

foreach my $doc (sort(keys(%DOC_TOTAL))){
#    print $_,"\t",
    my $correct=0;
    $correct=$DOC_CORRECT{$doc} if(defined($DOC_CORRECT{$doc}));
    my $total=$DOC_TOTAL{$doc};
#    if($total==0){
#	print "$doc\n";
#    }
    printf("%s %d / %d =\t%.5f\n",$doc,$correct,$total,($correct/$total));
}

print "\n";

print "TOTAL: $TOTAL\n";
print "CORRECT: $CORRECT\n";
printf("Accuracy: %.3f\n",(100*$CORRECT/$TOTAL));


sub isCorrect {
    my @left=@{(shift)};
    my @Q=@{(shift)};
    return 1 if($left[0]==-1 && scalar(@Q)==0);
    return 0 if($left[0]!=-1 && scalar(@Q)==0);
    return 0 if($left[0]==-1 && scalar(@Q)>0);
    foreach my $q (@Q){
	my ($gBeg,$gEnd,$gSen)=split(/,/,$q);
#	return 1 if($q->[0] == $left[0] && $q->[1] == $left[1]);
	return 1 if($gBeg == $left[0] && $gEnd == $left[1]);
    }
    return 0;
}


sub getClosestToTheLeft {
    my $line=shift;
    my $inv=shift;
    my $hashref=shift;
    if(defined($inv->{$line."_".$line})){
	my $CID=$inv->{$line."_".$line};
	my @cluster=@{$hashref->{$CID}};
	#find the leftmost
	my @left=(-1,-1,-1);
	foreach my $c(@cluster){
	    my ($leftBeg,$leftEnd,undef)=@{$c};
	    next if($leftBeg>$line); #next if to the right
	    next if($leftBeg==$line && $leftEnd==$line); #next if the same
	    if(defined($left[0]) && $left[0]<$leftBeg){
		@left=($leftBeg,$leftEnd,undef);
	    }
	}
	return @left;
    } else {
	return (-1,-1,-1);
    }
}

sub getInvertedHashRef {
    my $hr=shift;
    my $inv={};
    foreach my $CID (keys %{$hr}){
	foreach my $span(@{$hr->{$CID}}){
	    my ($beg,$end,$sent)=@{$span};
	    my $invKey=$beg."_".$end;
	    $inv->{$invKey}=$CID;
	}
    }
    return $inv;
}

sub getNextSys {
    my $header=<SYS>;
    chomp $header;
    return (undef,undef) if(!defined($header));
    my @doc;
    my $line;
    while($line=<SYS>){
	chomp $line;
	last if($line=~/^#end doc/);
	my @cols=split(/\s+/,$line);
	push @doc,\@cols;
    }
    my $hashref=Coref::getSpanHash(\@doc);
    my $inv=getInvertedHashRef($hashref);
    return ($header,$hashref,$inv);
}

sub getNextGold {
    my %x;
    return %x;
}
close GOLD;
