#!/bin/python
# -*- coding: utf-8 -*-

import sys
import codecs
import scipy.stats
import math
import argparse

sys.stdin = codecs.getreader('utf-8')(sys.stdin)
sys.stdout = codecs.getwriter('utf-8')(sys.stdout)

argpar = argparse.ArgumentParser(description='Performs two-sided t-test on related samples')
argpar.add_argument('-m1','--model1',dest='model1',help='the first model',required=True)
argpar.add_argument('-m2','--model2',dest='model2',help='the second model',required=True)
args = argpar.parse_args()


def ttest_rel( sample1, sample2 ):
  assert(len(sample1) == len(sample2))

  differences = map(lambda (x,y): x-y, zip(sample1,sample2))
  N = len(differences)
  mean = sum(differences)/N
  variance = sum(map(lambda x: pow(x-mean,2),differences))/(N-1) # N-1 for sample instead of whole group

  t = mean/math.sqrt(variance/N)
  
  p = scipy.stats.distributions.t.cdf(-abs(t),N-1)*2
  return t,p

sample1 = [ float(line.strip()) for line in codecs.open(args.model1,'r','utf-8') if line.strip() ]
sample2 = [ float(line.strip()) for line in codecs.open(args.model2,'r','utf-8') if line.strip() ]

if len(sample1) != len(sample2):
    print "sample sizes don't match - stopping"
    exit(1)

t,p = ttest_rel( sample1, sample2 )

print 'performing two-sided t-test on related samples'
print 'sample size =', len(sample1)
print 't =', t
print 'p-value =', p
