package ims.coref.util;

import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ims.coref.Options;
import ims.coref.data.Document;
import ims.coref.data.PairInstance;
import ims.coref.data.Span;
import ims.coref.features.AbstractSingleDataDrivenFeature;
import ims.coref.features.FeatureSet;
import ims.coref.features.IFeature;
import ims.coref.io.DocumentReader;
import ims.coref.io.ReaderWriterFactory;
import ims.coref.lang.Language;
import ims.coref.markables.IMarkableExtractor;
import ims.coref.markables.MarkableExtractorFactory;
import ims.coref.training.ITrainingExampleExtractor;
import ims.coref.training.TrainingExampleExtractorFactory;

public class FindFiringFeature {
	
	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		Language.initLanguage(options.lang);
		final IMarkableExtractor markableExtractor=MarkableExtractorFactory.getExtractorS(options.markableExtractors==null?Language.getLanguage().getDefaultMarkableExtractors():options.markableExtractors);
		final ITrainingExampleExtractor teex=TrainingExampleExtractorFactory.getExtractor(options.trainingExampleExtractor,markableExtractor,options.commitStrategy);
		AbstractSingleDataDrivenFeature[] features=getFeatures();
		DocumentReader reader=ReaderWriterFactory.getReader(options.inputFormat, options.input);
		
		Set<String> lookFor=getFeatureValuesToLookFor();
		
		if(markableExtractor.needsTraining()){
			System.out.println("Training anaphoricy classifier");
			markableExtractor.train(reader);
		}
		
		for(Document d:reader){
			List<PairInstance> instances=teex.getInstances(d);
			for(PairInstance pi:instances)
				for(AbstractSingleDataDrivenFeature f:features){
					String fVal=f.getStringValue(pi, d);
					if(lookFor.contains(fVal)){
//						String sameSen=pi.ant.s==pi.ana.s?"Y":"N";
//						String docNoBlanks=d.header.replaceAll(" ", "QQQ");
						String antKey=getKey(pi.ant);//docNoBlanks+"_"+pi.ant.s.sentenceIndex+"."+pi.ant.start+"."+pi.ant.hd+"."+pi.ant.end;
						String anaKey=getKey(pi.ana);//docNoBlanks+"_"+pi.ana.s.sentenceIndex+"."+pi.ana.start+"."+pi.ana.hd+"."+pi.ana.end;
						System.out.println(d.header.replaceAll(" ", "__") + "\t" + pi.corefers+"\t"+antKey+"-"+anaKey+"\t"+fVal);
					}
				}
		}
		
		
	}

	private static String getKey(Span ant) {
		return ant.s.sentenceIndex+"."+ant.start+"."+ant.hd+"."+ant.end;
	}

	private static Set<String> getFeatureValuesToLookFor() {
		String[] values={
//				"nsubj^dep_++you",
//				"dobj^conj^dobj_++it",
//				"nsubj^advcl^prep_pobj_appos_++you",
//				"dobj^nsubj_++itself",
				
				//Disref below
				"dobj^nsubj_++him",
				"pobj^prep^nsubj_++him",
				"dobj^nsubj_++them",
				"dobj^nsubj_++it"
		};
		Set<String> r=new HashSet<String>();
		Collections.addAll(r, values);
		return r;
	}

	private static AbstractSingleDataDrivenFeature[] getFeatures() {
		String[] names=new String[]{
				"SSPathFun+AnaphorPronounForm"
		};
		FeatureSet fs=FeatureSet.getFromNameArray(names);
		List<IFeature> iF=fs.getFeatures();
		AbstractSingleDataDrivenFeature[] r=new AbstractSingleDataDrivenFeature[iF.size()];
		for(int i=0;i<iF.size();++i){
			r[i]=(AbstractSingleDataDrivenFeature) iF.get(i);
		}
		return r;
	}
}
